#if !defined(AFX_DMD_H__0169C360_5BDA_499F_89BA_3FE7FCB29D09__INCLUDED_)
#define AFX_DMD_H__0169C360_5BDA_499F_89BA_3FE7FCB29D09__INCLUDED_

#if _MSC_VER >= 1000
#pragma once
#endif // _MSC_VER >= 1000
// DMD.h : header file
//

#define DMD_ROWS         32
#define DMD_COLUMNS     128
#define DMD_PAGE_BYTES  ((DMD_ROWS * DMD_COLUMNS)/8)  // 512 bytes

#define PAGE_LENGTH  0x4000

#define WRITE_TYPE_COLUMNS       0
#define WRITE_TYPE_ROWS          1

#define PIXEL_HEIGHT  5
#define PIXEL_WIDTH   5

#define MAX_GRAPHIC_INDEX  2048 // at this time we will restrict value

#define PLANE0_ON      0x01
#define PLANE0_XORED   0x02
#define PLANE0_SKIPPED 0x04
#define PLANE1_ON      0x08
#define PLANE1_XORED   0x10
#define PLANE1_SKIPPED 0x20

#define PLANE_STATUS_VALID                 0
#define PLANE_STATUS_UNKNOWN_TYPE          1
#define PLANE_STATUS_UNIMPLEMENTED_TYPE    2
#define PLANE_STATUS_TABLEENTRYOUTOFRANGE  3
#define PLANE_STATUS_IMAGEOUTOFRANGE       4

#define DEFAULT_DEBOUNCE_TIME  500  // 1/2 second before repeating graphics quickly
#define DEBOUNCE_STATE_IDLE        0
#define DEBOUNCE_STATE_DEBOUNCING  1
#define DEBOUNCE_STATE_DONE        2

const static char StatusText [(PLANE_STATUS_IMAGEOUTOFRANGE + 1)][16] = 
	{
		"Valid Decoding",
		"Invalid (1)",
		"Invalid (2)",
		"Invalid (3)",
		"Invalid (4)"
	};



/////////////////////////////////////////////////////////////////////////////
// DMD dialog

struct DMDData
{
	unsigned long ROMSize;
	unsigned long TableAddress;
	int CurrentIndex;
	unsigned char Plane_0Status;
	unsigned char Plane_0[DMD_PAGE_BYTES];
	unsigned char Plane_0Skipped[DMD_PAGE_BYTES];
	unsigned char Plane_0Xored[DMD_PAGE_BYTES];
	unsigned int Plane_0Size;
	unsigned char Plane_1Status;
	unsigned char Plane_1[DMD_PAGE_BYTES];
	unsigned char Plane_1Skipped[DMD_PAGE_BYTES];
	unsigned char Plane_1Xored[DMD_PAGE_BYTES];
	unsigned int Plane_1Size;
};

struct ImageHeader
{
	unsigned char SpecialFlagByte;
	unsigned char RepeatBytes[8];
	unsigned char ReadMask;
};

class DMD : public CDialog
{
// Construction
public:
//	DMD(CWnd* pParent = NULL);   // standard constructor
	DMD(CWnd* pParent = NULL, RegistrySettings *mainptr = NULL);   // standard constructor
	~DMD();
	BOOL Create();
	unsigned char Init();
	DMDData Data;
	void DecodeImageCurrentIndex(void);
	void DecodeImage(int Index, int Plane);
//	unsigned char DecodeImage(unsigned char *Source, unsigned char *Dest, unsigned char *Skipped, unsigned char *Xored);
	unsigned char DecodeImage(unsigned char **Source, unsigned char *Dest, unsigned char *Skipped, unsigned char *Xored);

//	void Decode_00(unsigned char *Source, unsigned char *Dest);
	void Decode_00(unsigned char **Source, unsigned char *Dest);
//	void Decode_01(unsigned char *Source, unsigned char *Dest);
	void Decode_01(unsigned char **Source, unsigned char *Dest);
//	void Decode_02(unsigned char *Source, unsigned char *Dest);
	void Decode_02(unsigned char **Source, unsigned char *Dest);
//	void Decode_03(unsigned char *Source, unsigned char *Dest);
	void Decode_03(unsigned char **Source, unsigned char *Dest);
//	void Decode_04(unsigned char *Source, unsigned char *Dest);
	void Decode_04(unsigned char **Source, unsigned char *Dest);
//	void Decode_05(unsigned char *Source, unsigned char *Dest);
	void Decode_05(unsigned char **Source, unsigned char *Dest);
//	void Decode_06(unsigned char *Source, unsigned char *Dest, unsigned char *Xored);
	void Decode_06(unsigned char **Source, unsigned char *Dest, unsigned char *Xored);
//	void Decode_07(unsigned char *Source, unsigned char *Dest, unsigned char *Xored);
	void Decode_07(unsigned char **Source, unsigned char *Dest, unsigned char *Xored);
//	void Decode_08(unsigned char *Source, unsigned char *Dest, unsigned char *Skipped);
	void Decode_08(unsigned char **Source, unsigned char *Dest, unsigned char *Skipped);
//	void Decode_09(unsigned char *Source, unsigned char *Dest, unsigned char *Skipped);
	void Decode_09(unsigned char **Source, unsigned char *Dest, unsigned char *Skipped);
//	void Decode_0A(unsigned char *Source, unsigned char *Dest, unsigned char *Skipped);
	void Decode_0A(unsigned char **Source, unsigned char *Dest, unsigned char *Skipped);
//	void Decode_0B(unsigned char *Source, unsigned char *Dest, unsigned char *Skipped);
	void Decode_0B(unsigned char **Source, unsigned char *Dest, unsigned char *Skipped);
	unsigned char ReadNext8BitValue(struct ImageHeader *Header, unsigned char **SourcePtr);
	unsigned char ReadNextBit(struct ImageHeader *Header, unsigned char **SourcePtr);
	void WriteNext8BitValue(unsigned char **DestPtr, unsigned int *WriteCounterPtr,unsigned char ch, unsigned char Type);
	void Decode_01or02(unsigned char **SourcePtr, unsigned char **DestPtr, unsigned char Type);
	void Decode_04or05(unsigned char **SourcePtr, unsigned char **DestPtr, unsigned char Type);
	void Decode_06or07(unsigned char **SourcePtr, unsigned char **DestPtr, unsigned char **XoredPtr, unsigned char Type);
	void Decode_08or09(unsigned char **SourcePtr, unsigned char **DestPtr, unsigned char **Skipped, unsigned char Type);
	void Decode_0Aor0B(unsigned char **SourcePtr, unsigned char **DestPtr, unsigned char **Skipped, unsigned char Type);

	void InvalidateDMDPages();
	void UpdateStaticTextBoxes();
	void UpdateControls();
	void SaveDMDRegistrySettings();
	void UpdateCheckboxText();

// Dialog Data
	//{{AFX_DATA(DMD)
	enum { IDD = IDD_DMD };
	CSpinButtonCtrl	m_Spin;
	CListBox	m_PixelColor;
	CButton	m_Xored;
	CButton	m_Skipped;
	CStatic	m_Both;
	CStatic	m_Medium;
	CStatic	m_Dim;
	CButton	m_NextGraphic;
	CButton	m_PreviousGraphic;
	CStatic	m_UserReg;
	//}}AFX_DATA


// Overrides
	// ClassWizard generated virtual function overrides
	//{{AFX_VIRTUAL(DMD)
	public:
	virtual BOOL Create(LPCTSTR lpszClassName, LPCTSTR lpszWindowName, DWORD dwStyle, const RECT& rect, CWnd* pParentWnd, UINT nID, CCreateContext* pContext = NULL);
	protected:
	virtual void DoDataExchange(CDataExchange* pDX);    // DDX/DDV support
	//}}AFX_VIRTUAL

// Implementation
protected:
	RegistrySettings *PassedInPointer;
//    CBrush* m_pListBkBrush;
#if ALLOW_MOUSE_TO_REPEAT
	int TimerIDDMD;
	unsigned int NextDebounce;
	unsigned int PreviousDebounce;
	unsigned char NextDebounceState;
	unsigned char PreviousDebounceState;
#endif
	unsigned char NaggedOnce;
//	unsigned char NextPressed;
//	unsigned char PreviousPressed;
	// Generated message map functions
	//{{AFX_MSG(DMD)
	afx_msg int OnCreate(LPCREATESTRUCT lpCreateStruct);
	afx_msg void OnDestroy();
	virtual BOOL OnInitDialog();
	afx_msg void OnPaint();
	afx_msg void OnButtonNextGraphic();
	afx_msg void OnButtonPreviousGraphic();
	afx_msg void OnCheckSkipped();
	afx_msg void OnCheckXored();
	afx_msg void OnSelchangeList1();
	virtual void OnOK();
	afx_msg void OnClose();
	afx_msg HBRUSH OnCtlColor(CDC* pDC, CWnd* pWnd, UINT nCtlColor);
	afx_msg void OnShowWindow(BOOL bShow, UINT nStatus);
	afx_msg void OnTimer(UINT nIDEvent);
	afx_msg void OnLButtonDown(UINT nFlags, CPoint point);
	afx_msg void OnLButtonUp(UINT nFlags, CPoint point);
	//}}AFX_MSG
	DECLARE_MESSAGE_MAP()

};

//{{AFX_INSERT_LOCATION}}
// Microsoft Developer Studio will insert additional declarations immediately before the previous line.

#endif // !defined(AFX_DMD_H__0169C360_5BDA_499F_89BA_3FE7FCB29D09__INCLUDED_)
