/***************************************************/
/* IJCSUM                                          */
/*                                                 */
/* A checksum utility for hacking Indiana Jones    */
/* pinball ROM images.                             */
/*                                                 */
/* This utility was developed for private owners   */
/* of Williams' Indiana Jones pinball machine.     */
/* It is written in no way to profit from, or to   */
/* defraud the excellent work by the Indiana Jones */
/* design team.  When hacking any pinball ROM,     */
/* please do not edit, modify, delete or change    */
/* any of the text which contains the credits to   */
/* to the original Indiana Jones software          */
/* designers.                                      */
/*                                                 */
/* This utility is designed for educational        */
/* purposes only.  My main goal was to edit text   */
/* within the ROM image to change what is          */
/* displayed on the pinball display.  In the       */
/* future, I may disassemble the code and figure   */
/* out where such bytes control certain game       */
/* timers and features.                            */
/*                                                 */
/* Revision History:                               */
/* 1.0  April 13, 1999  Initial coding of program  */
/*      using Microsoft Visual C++ 5.0             */
/*                                                 */
/* 2.0  November 23, 1999 After using program      */
/*      a few times I realize I didn't quite get   */
/*      the program to write the correct checksum  */
/*      to the output file all the time.           */
/*      Problem should be fixed now.               */
/*                                                 */
/* 3.0  December 21, 2002.                         */
/*      Changing from MS C++ to Borland. Added     */
/*      customised version number (2nd byte of     */
/*      CRC is version number x in L-x).           */
/*                                                 */
/***************************************************/

//----------------------------------------------------------------------------//
//
// Notes regarding version numbers
//
// The 2nd byte of Checksum at 0x7FFEF represents the "version" number that is
// displayed on the dot-matrix during power-up and when entering the menus.
//
// When the value of the version-byte is 0xD0 - 0xDF, the version will be
// considered Prototype and the text string from ROM "REV. P-" will be prepended
// to the version number, and the version number will have 0xD0 subtraced from
// it.
//
// This means that version-byte of 0xD0 will show "REV. P-0"
// This means that version-byte of 0xFF will show "REV. P-2F"
//
// When the value of the version-byte is 0x00 - 0xCF, the version will be
// considered production and the text string from ROM "REV. L-" will be
// prepended to the version number.
//
// This means that version-byte of 0x00 will show "REV. L-0"
// This means that version-byte of 0xCF will show "REF. L-CF"
//
// Text is always right-justified on the display (for 1 and 2 digit versions).
//
//----------------------------------------------------------------------------//
//
// IJ has Version Text at the following locations (in domestic L-7)
//
// 11F7C "REV. LC-%"  // coin editor version?
// 11F88 "REV. PSD%"  // coin editor version?
// 11FED "REV. LG-%"  // Not sure
// 6C70E "REV. L-%    // For export machines, this is the main L-x string
// 6C71E "REV. P-%    // For export machines, this is the main L-x string
// 70DBF "REV. L-%    // This is the main rev used for version-bytes 0x00-0xCF
// 70DCA "REV. P-%    // This is the secondary string for bytes 0xD0-0xFF
//
//----------------------------------------------------------------------------//
//
// For public-release (underground) Hack roms, a primary person should be the
// "keeper" of the version numbers, this way there is some sort of orginization.
// When somebody wants to "release" an image they should email the keeper of the
// particular game and ask what version number to assign to it.  The keeper can
// record THAT version number for the particular release and forever-on that
// version number will be associated with whatever functionality the hacker has
// put into the rom.
//
//
// At this time, for Indiana Jones.  The version strings in ROM will be modified
// with HK string to indicate a Hack revision.  The string at memory locations
// 6C70E and 70DBF will be changed from "REV. L-" to "REV. HK".  The first
// public (underground) release will be HK1, the version-byte will have the
// value of 0x01.
//
// A record of version-functionality will be kept in a different file and
// probably saved online.
//
//----------------------------------------------------------------------------//
//
#include <string.h>  //string.h has strlen() and strcpy() functions I use.
#include <iomanip.h> //iomanip.h has setiosflags() and resetiosflags()
#include <fstream.h> //fstream.h contains fstream file object. File manipulation
#include <stdlib.h>  //stdlib.h has exit() function I use.
#include <ctype.h>   // This has toupper()


#define LARGEFILES (0)



#define BYTE unsigned char      // 8-bit memory storage
#define UINT unsigned int       // 16-bit memory storage
#define ULNG unsigned long      // 32-bit memory storate


// Defines used for global 'Flags' byte
#define FLAG_VERSION_SPECIFIED    0x01
#define FLAG_FILE_SPECIFIED       0x02

#if LARGEFILES
const ULNG FileSize = 1048576L;  //Number of bytes the Input file is expected to be
#else
const ULNG FileSize = 524288L;  //Number of bytes the Input file is expected to be
#endif

//Below are the function prototypes.  Actual function code is further down.
static void Disclaimer(void);
//static BYTE OpenFile(fstream &Fin, int ArgsNum, char **Args, int code);
static BYTE OpenFile(fstream &Fin);
static BYTE VerifyInFile(fstream &Fin);
static BYTE MakeBackup(fstream &Fin);
static BYTE ZeroCsum(fstream &Fin, UINT *csum, BYTE *ByteZero);
static BYTE CalcCsum(UINT *csum, int *extra);
static BYTE WriteCsum(fstream &Fin, UINT &csum, int extra, BYTE ByteZero);
static void PreProcessCLArgs(int ArgsNum, char **Args);
//End of function prototypes

// Below are global variables
char FileName[255];    // Filename to fiddle with
BYTE Version;          // Desired Version
BYTE Flags;            // Misc flags
// End of global variables

void main(int ArgsNum, char **Args)
{
	fstream Fin;
	UINT csum;
	BYTE ByteZero = 0;
	int extra = 0;

	PreProcessCLArgs(ArgsNum, Args);

	Disclaimer();  //Display the 'disclaimer' to the user
	cout << "\n\nOpening the input file...";
	if (!OpenFile(Fin))
	{
		if (!(Flags & FLAG_FILE_SPECIFIED))
		{
			cerr << "Error!\n\nCould not open default file IJHACK.ROM."
				 << "\nMake sure it's in the same direcory as IJCSUM.EXE.\n\n";
			exit(1);
		}
		cerr << "Error!\n\nCould not open user specified file: "
			 << Args[1]
			 << "\n";
		exit(1);
	}
	cout << "OK"
		 << "\nChecking the file size..."
		 << flush;
	if (!VerifyInFile(Fin))
	{
		cerr << "Error!\n\nThe file failed the size test.  It must be "
			 << FileSize
			 << " bytes long.";
		exit(1);
	}
	cout << "OK"
		 << "\nCreating a backup copy..."
		 << flush;
	Fin.close();
	OpenFile(Fin);
	if (!MakeBackup(Fin))
	{
		cerr << "Warning!\n\nUnable to create backup copy.  This error message\n"
			 << "should not occur.  Possibly out of memory or disk space.\n"
			 << "Attempting to continue...";
	}
	cout << "OK"
		 << "\nCalculating checksum....."
		 << flush;
	Fin.close();
	OpenFile(Fin);
	if (!ZeroCsum(Fin, &csum, &ByteZero))
	{
		cerr << "Error\n\nUnexpected error in 'ZeroCsum' routine.";
		exit(1);
	}
	if (!CalcCsum(&csum, &extra))
	{
		cerr << "Error\n\nUnable to calculate ANY valid checksum.\n"
			 << "This error should not ever happen, program bug.\n\n";
		exit(1);
	}
	Fin.close();
	OpenFile(Fin);
	if (!WriteCsum(Fin, csum, extra, ByteZero))
	{
		cerr << "Error\n\nUnexpected error in 'WriteCsum' routine.";
		exit(1);
	}
	cout << "Success!\n\n"
		 << "A CheckSum of: 0x"
		 << setiosflags(ios::uppercase)
		 << hex << csum
		 << " was written to the input file at\n"
		 << "the checksum memory location 7FFEE and 7FFEF!\n\n";
	if (extra > 0)
	{
		cout << "In order to generate a valid checksum, it was necessary\n"
			 << "to change a third byte of the ROM.  Byte 0x00000 had to\n"
			 << "be incremented by 0x" << hex << extra
			 << " in order to generate the above\n"
			 << "checksum value.  Byte 0x00000 in Indiana Jones ROM is\n"
			 << "only used to store text that isn't used in the program\n"
			 << "so modifying this byte will not affect the game play in\n"
			 << "any way.\n";
	}

	cout << "\nPress <ENTER> to return to the operating system...";
	cin.ignore(255,'\n');
}




static void Disclaimer(void)
{
	cout << "\n\nIJCSUM   v1.0\n\n"
		 << "A checksum utility for hacking Indiana Jones pinball ROM images.\n\n"
		 << "This utility was developed for private owners of Williams' Indiana\n"
		 << "Jones pinball machine.  It is written in no way to profit from, or\n"
		 << "to defraud the excellent work by the Indiana Jones design team.\n"
		 << "When hacking any pinball ROM, please do not edit, modify, delete or\n"
		 << "change any of the text which contains the credits to to the original\n"
		 << "Indiana Jones software designers.\n\n"
			<< "This utility is designed for educational purposes only.  My main goal\n"
		 << "was to edit text within the ROM image to change what is displayed on\n"
		 << "the pinball display.  In the future, I may disassemble the code and\n"
		 << "figure out where such bytes control certain game timers and features.\n\n"
		 << "Please be aware that if WMS gets the slightest hint that you are\n"
		 << "modifying their ROMs, they will banish you from their ROM library and\n"
		 << "will not be able to download ROMs from their web site!\n\n"
		 << "Press <ENTER> to continue...";
	cin.ignore(255,'\n');
}


static BYTE OpenFile(fstream &Fin)
{
	static BYTE OneTime = 0;
	BYTE OpenOkay;

	Fin.clear(0);
	Fin.open(FileName, ios::binary|ios::in|ios::out|ios::nocreate, filebuf::openprot);
	OpenOkay = 0x00;
	if (!(Fin.fail()))
	{
		int i = strlen(FileName);
		OpenOkay++;
		if (!OneTime)
		{
			cout << "Opened " << FileName;
			if (i < 22)
				for (int j = 26 - i; j > 0; j--)
					cout << '.';
			OneTime++;
		}
	}
	Fin << resetiosflags(ios::skipws);
	Fin >> resetiosflags(ios::skipws);
	return OpenOkay;
}



static BYTE VerifyInFile(fstream &Fin)
{
	char Buffer;
	long counter = 0;

	Fin >> resetiosflags(ios::skipws);
	Fin >> Buffer;
	while(!Fin.eof())
	{
		counter++;
		Fin >> Buffer;
	}
	if (counter == FileSize)
	{
		cout << "File is " << counter << " bytes.............";
		return 0x01;
	}
	return 0x00;
}

static BYTE MakeBackup(fstream &Fin)
{
	char Buffer;
	ofstream Fout;
	BYTE GoodFlag;
	char FName[80];
	long i;
	long Dot;

	Dot = 0;
	strcpy(FName, FileName);

	for (i = 0; FName[(int)i]; i++)
		if (FName[(int)i] == '.')
			Dot = i;

	if (Dot)
		i = Dot;

	FName[(int)(i++)] = '.';
	FName[(int)(i++)] = 'B';
	FName[(int)(i++)] = 'A';
	FName[(int)(i++)] = 'K';
	FName[(int)i] = 0x00;

	Fout.clear(0);
	Fout.open(FName,ios::binary);
	Fout << resetiosflags(ios::skipws);
	GoodFlag = 0x00;
	if (!(Fout.fail()))
	{
		GoodFlag++;
		cout << "File backed up to " << FName;
		i = strlen(FName);
		if (i < 13)
			for(int j = 15 - (int)i; j > 0; j--)
				cout << '.';
	}

	if (GoodFlag)
	{
		Fin >> resetiosflags(ios::skipws);
		for (i = 0; i < FileSize; i++)
		{
			Fin >> Buffer;
			Fout << Buffer;
		}
	}
	return GoodFlag;
}


static BYTE ZeroCsum(fstream &Fin, UINT *csum, BYTE *ByteZero)
{
	register UINT Local_csum = 0;
	BYTE GoodFlag = 0;
	BYTE Buffer;
	long i;

	for (i = 0; i < FileSize; i++)
	{
		switch (i)
		{
#if LARGEFILES
			case 0xFFFEEL :
#else
			case 0x7FFEEL :
#endif
				Fin << 0x00;
				GoodFlag++;
				break;

#if LARGEFILES
			case 0xFFFEFL :
#else
			case 0x7FFEFL :
#endif
				if (GoodFlag)
					Fin << 0x00;
				break;

			case 0x00000L :
				Fin >> *ByteZero;
				Local_csum += *ByteZero;
				break;

			default :
				Fin >> Buffer;
				Local_csum += Buffer;
		}
	}
	*csum = Local_csum;
	return GoodFlag;
}


static BYTE CalcCsum(UINT *csum, int *extra)
{
	register int x,y,z;
	UINT temp;
	UINT hit;
	BYTE stop = 0;
	BYTE StartYValue = 0x00;
	BYTE EndYValue = 0xFF;

	if (Flags & FLAG_VERSION_SPECIFIED)
		StartYValue = EndYValue = Version;

	for (x = 0; ((x <= 0xff) && !stop); x++)
	{
		temp = *csum + x;
		for (y = StartYValue; ((y <= EndYValue) && !stop); y++)
		{
			hit = y + temp;
			if (hit == (((UINT)x << 8) | y))
				stop++;
		}
	}
	if (!stop)
	{
		for(z = 0;(( z<= 0xff) && !stop); z++)
		{
			for(x = 0; ((x <= 0xff) && !stop); x++)
			{
				temp = *csum + x + z;
				for(y = StartYValue; ((y <= EndYValue) && !stop); y++)
				{
					hit = y + temp;
					if (hit == (((UINT)x << 8) | y))
						stop++;
				}
			}
		}
		*extra = (z - 1);
	}
	*csum = hit;
	return stop;
}


static BYTE WriteCsum(fstream &Fin, UINT &csum, int extra, BYTE ByteZero)
{
	BYTE GoodFlag = 0x00;
	BYTE Buffer;
	long i;
	char csumByte1;
	char csumByte2;
	char extrabyte=NULL;
	int offset = 0;

	csumByte2 = (csum & 0x00FF);
	csumByte1 = ((csum & 0xFF00) >> 8);

	if (extra > 0)
		extrabyte = extra + ByteZero;

	for (i = 0; i < FileSize; i++)
	{
		if (!i && extrabyte)
		{
			Fin << extrabyte;
			offset=1;
		}
#if LARGEFILES
		if (i == 0xFFFEEL - offset)
#else
		if (i == 0x7FFEEL - offset)
#endif
		{
			Fin << csumByte1;
			GoodFlag++;
		}
#if LARGEFILES
		else if (i == 0xFFFEFL - offset)
#else
		else if (i == 0x7FFEFL - offset)
#endif
		{
			if (GoodFlag)
				Fin << csumByte2;
		}
		else
		{
			Fin >> Buffer;
		}
	}
	if (GoodFlag) cout << "Modified the input file!.........";
	return GoodFlag;
}

static void PreProcessCLArgs(int ArgsNum, char **Args)
{
	BYTE ch;

	Version     =
	FileName[0] =
	Flags       = 0x00;

	while (--ArgsNum)
	{
		ch = *Args[ArgsNum];
		switch (ch)
		{
			case '-'  :  // start of an option
				ch = *(Args[ArgsNum] + 1);
				switch (toupper(ch))
				{
					case 'V' :  // Desired pinball rom version specified.
						ch = toupper(*(Args[ArgsNum] + 2));
						if (ch)
						{
							ch -= 0x30;
							if (ch > 9)
								ch -= 7;
							Version = ch;
							ch = toupper(*(Args[ArgsNum] + 3));
							if (ch)
							{
								ch -= 0x30;
								if (ch > 9)
									ch -= 7;
								Version = (Version << 4) | ch;
							}
							Flags |= FLAG_VERSION_SPECIFIED;
						}
						break;

					default :
						break;
				}
				break;

			case 0x00 :
				break;

			default  :  // otherwise, a text string will be considered to be the
							// filename.
				strcpy (FileName, Args[ArgsNum]);
				Flags |= FLAG_FILE_SPECIFIED;
		}
	}
	if (!(Flags & FLAG_FILE_SPECIFIED))
		strcpy (FileName, "IJHACK.ROM");
}
